#include "stdio.h"
#include "stdlib.h"
#include "string.h"
#include "stm8s.h"
#include "usrlib-uart.h"
#include "usrlib-sd.h"

char UARTstrTX[80]; // USRT send buffer

void UART_Menu(void);

void main(void)
{
	CLK_ClockSwitchConfig(CLK_SWITCHMODE_AUTO, CLK_SOURCE_HSI, DISABLE, CLK_CURRENTCLOCKSTATE_DISABLE);
	CLK_HSIPrescalerConfig( CLK_PRESCALER_HSIDIV1 );
	CLK_SYSCLKConfig( CLK_PRESCALER_CPUDIV1 );
	
	UART2_DeInit();
  UART2_Init((u32)115200, UART2_WORDLENGTH_9D, UART2_STOPBITS_1, UART2_PARITY_ODD, UART2_SYNCMODE_CLOCK_DISABLE, UART2_MODE_TXRX_ENABLE);	
  UART2_ITConfig(UART2_IT_RXNE_OR, ENABLE);
	UART2_Cmd(ENABLE);

  // LDE indicator
	GPIO_Init(GPIOD, 0b00000001, GPIO_MODE_OUT_PP_HIGH_FAST);

	enableInterrupts();
	
	SerialPutString("\r\nstart SD card IF test\r\n");

	// SD card detect pin pullup
	GPIO_Init(SD_DETECT_GPIO_PORT, SD_DETECT_PIN, GPIO_MODE_IN_PU_NO_IT);

	// Wait for SD card insertion
	SerialPutString("waiting for SD card\r\n");
	while(SD_Detect()==SD_NOT_PRESENT){;}
	SerialPutString("SD card inserted\r\n");

  SD_Init(); // Initialization SPI and SD card

	while(1) { UART_Menu(); } // main loop
}

void DumpBlock(void)
{
	int i,k;
	for(i=0;i<SD_BLOCK_SIZE;i=i+16)
	{
		char ascii[3], str[17];
		UARTstrTX[0]=0; // null string
		str[16]=0;
		for(k=0;k<=15;k++)
		{
			sprintf(ascii,"%02x ",SDstr[i+k]);
			strcat(UARTstrTX,ascii);
			str[k]=SDstr[i+k];
			if(str[k]<32)str[k]=46;
			if(str[k]==0x7F)str[k]=46;
			if((str[k]&0xF0)==0x80)str[k]=46;
			if((str[k]&0xF0)==0x90)str[k]=46;
			if((str[k]&0xF0)==0xE0)str[k]=46;
			if((str[k]&0xF0)==0xF0)str[k]=46;
		}
		strcat(UARTstrTX," | ");
		strcat(UARTstrTX,str);
		strcat(UARTstrTX,"\r\n");
		SerialPutString(UARTstrTX);
	}
}

// generate incremental dummy data to save a log file
// x : BYTE of generating data
u32 SD_size_byte; // 0-2GB
u8 SD_full; // SD card full flag
void log_dummy(u32 x, char v) {
	u32 i;
	SD_write_sequencial_init();
	for(i=0;i<SD_BLOCK_SIZE;i++) SDstr[i]=0; // clr work area
	for(i=0;i<(x>>3);i++) {
		char c[12];
		sprintf(c,"%04x%04x",(u16)((i&0xFFFF0000)>>16),(u16)(i&0xFFFF));
		SD_write_sequencial(c,v);
	}
	SD_write_sequencial("",v); // flush data
}

void UART_Menu(void)
{
	uint32_t i;
	if(UART_STR_EXIST==1) // there is a command line
	{
		int fieldnumber;
		uint32_t address;

		fieldnumber = separate_line(command); // divide a line into words

		// command parse
		//  inc 2           ---> write increment data to block 2 (address 2*512)
		//  fill 49 129     ---> write 49 to block 129 (address 129*512)
		//  fill 55 12 133  ---> write 55 to block 12-133
		//  txt abc456 1996 ---> write "abc456" to block 1996 (address 1996*512)
		//  format          ---> format 2GB FAT16
		//  log 10000 v     ---> genrate log file 10000BYTE (v means vervose)
		//  3               ---> dump block 3 (address 3*512)
		//  3 18            ---> dump block from 3 to 18
		if(strcmp(Field[0],"inc")==0 && fieldnumber==2)
		{
			address=(uint32_t)(atof(Field[1])*SD_BLOCK_SIZE);
			for(i=0;i<SD_BLOCK_SIZE;i++) SDstr[i]=(uint8_t)i; // set Tx data
			SD_WriteBlock(SDstr, address);	// Write block of 512 bytes
		}
		else if(strcmp(Field[0],"fill")==0 && fieldnumber==3)
		{
			uint8_t c;
			c=(uint8_t)atof(Field[1]);
			address=(uint32_t)(atof(Field[2])*SD_BLOCK_SIZE);
			for(i=0;i<SD_BLOCK_SIZE;i++) SDstr[i]=c; // set Tx data
			SD_WriteBlock(SDstr, address);	// Write block of 512 bytes
		}
		else if(strcmp(Field[0],"fill")==0 && fieldnumber==4)
		{
			uint8_t c;
			uint32_t block_stt, block_end, block;
			c=(uint8_t)atof(Field[1]);
			block_stt=(uint32_t)atof(Field[2]);
			block_end=(uint32_t)atof(Field[3]);
			for(i=0;i<SD_BLOCK_SIZE;i++) SDstr[i]=c; // set Tx data
			SerialPutString("\r\n");
			for(block=block_stt;block<=block_end;block++)
			{
				if((block % 6)==5)SerialPutString("\r\n");
				address=(uint32_t)(block*SD_BLOCK_SIZE);
				SD_WriteBlock(SDstr, address);	// Write block of 512 bytes
				sprintf(UARTstrTX,"0x%04x%04x  ",(int)(block>>16),(int)(block&0xFFFF));
				SerialPutString(UARTstrTX);
			}
		}
		else if(strcmp(Field[0],"txt")==0 && fieldnumber==3)
		{
			for(i=0;i<SD_BLOCK_SIZE;i++) SDstr[i]=0xFF; // set Tx data
			strcpy(SDstr,Field[1]); // set Tx data
			address=(uint32_t)(atof(Field[2])*SD_BLOCK_SIZE);
			SD_WriteBlock(SDstr, address);	// Write block of 512 bytes
		}
		else if(strcmp(Field[0],"format")==0 && fieldnumber==1)
		{
			SerialPutString("\r\n");
			SD_format_2GB();
		}
		else if(strcmp(Field[0],"log")==0 && (fieldnumber==2||fieldnumber==3))
		{
			u32 n;
			n=(u32)(atof(Field[1]));
			SerialPutString("\r\n");
			if(fieldnumber==2) log_dummy(n,' ');
			if(fieldnumber==3) log_dummy(n,Field[2][0]);
		}
		else if(fieldnumber==1) // dump a block
		{
			address=(uint32_t)(atof(Field[0])*SD_BLOCK_SIZE);
			if(address==0 && Field[0][0]!=0x30) goto NG; // adrs 0 but not "0"
			sprintf(UARTstrTX,"\r\nblock 0x%04x%04x\r\n",(int)(((uint32_t)atof(Field[0]))>>16),(int)(((uint32_t)atof(Field[0]))&0xFFFF));
			SerialPutString(UARTstrTX);
			SD_ReadBlock(SDstr, address); // Read block of 512 bytes
			DumpBlock();
		}
		else if(fieldnumber==2) // dump blocks
		{
			uint32_t block_stt, block_end;
			block_stt=(uint32_t)atof(Field[0]);
			block_end=(uint32_t)atof(Field[1]);
			if(block_stt==block_end) goto NG;
			if(block_stt>block_end) goto NG;
			for(i=block_stt;i<=block_end;i++)
			{
				address=(uint32_t)(i*SD_BLOCK_SIZE);
				sprintf(UARTstrTX,"\r\nblock 0x%04x%04x\r\n",(int)((uint32_t)i>>16),(int)((uint32_t)i&0xFFFF));
				SerialPutString(UARTstrTX);
				SD_ReadBlock(SDstr, address); // Read block of 512 bytes
				DumpBlock();
			}
		}
		else
		{
NG:
			SerialPutString("\r\n\nCOMMAND NG!\r\n");
		}

MENU:
		SerialPutString("\r\n");
		SerialPutString("inc 2         : write incremental to block 2 (adrs=2*512)\r\n");
		SerialPutString("fill 49 129   : write 49 to block 129 (adrs=129*512)\r\n");
		SerialPutString("fill 49 12 129: write 49 to block 12-129\r\n");
		SerialPutString("txt hello 996 : write hello to block 996 (adrs=996*512)\r\n");
		SerialPutString("format        : format 2GB FAT16\r\n");
		SerialPutString("log 10000 v   : genrate log file 10000BYTE (v:verbose)\r\n");
		SerialPutString("3             : dump block 3 (adrs=3*512)\r\n");
		SerialPutString("3 18          : dump block from 3 to 18\r\n");
		UART_STR_EXIST=0;
	}
	return;
}

#ifdef USE_FULL_ASSERT
void assert_failed(uint8_t* file, uint32_t line)
{ 
  while (1)  {  }
}
#endif
