#include "stdio.h"
#include "stdlib.h"
#include "string.h"
#include "stm8s.h"
#include "usrlib-uart.h"
#include "usrlib-sd.h"

// SD card format 2GB
uint8_t SDstr[SD_BLOCK_SIZE]; // SD card read/write buffer 512bytes
void SD_format_2GB(void)
{
	int i;
// BPB (block 0-5)
	for(i=0;i<SD_BLOCK_SIZE;i++) SDstr[i]=0;
	SDstr[0] =0xeb; // BLOCK 0
	SDstr[1] =0x3c;
	SDstr[2] =0x90;
	SDstr[3] =0x4d;
	SDstr[4] =0x53;
	SDstr[5] =0x44;
	SDstr[6] =0x4f;
	SDstr[7] =0x53;
	SDstr[8] =0x35;
	SDstr[9] =0x2e;
	SDstr[10]=0x30;
	SDstr[11]=0x00;
	SDstr[12]=0x02;
	SDstr[13]=0x40;
	SDstr[14]=0x06;
	SDstr[15]=0x00;
	SDstr[16]=0x02;
	SDstr[17]=0x00;
	SDstr[18]=0x02;
	SDstr[19]=0x00;
	SDstr[20]=0x00;
	SDstr[21]=0xf8;
	SDstr[22]=0xe9;
	SDstr[23]=0x00;
	SDstr[24]=0x3f;
	SDstr[25]=0x00;
	SDstr[26]=0xff;
	SDstr[27]=0x00;
	SDstr[28]=0x00;
	SDstr[29]=0x00;
	SDstr[30]=0x00;
	SDstr[31]=0x00;
	SDstr[32]=0x00;
	SDstr[33]=0x10;
	SDstr[34]=0x3a;
	SDstr[35]=0x00;
	SDstr[36]=0x80;
	SDstr[37]=0x00;
	SDstr[38]=0x29;
	SDstr[39]=0x29;
	SDstr[40]=0xf0;
	SDstr[41]=0x4a;
	SDstr[42]=0xec;
	SDstr[43]=0x4e;
	SDstr[44]=0x4f;
	SDstr[45]=0x20;
	SDstr[46]=0x4e;
	SDstr[47]=0x41;
	SDstr[48]=0x4d;
	SDstr[49]=0x45;
	SDstr[50]=0x20;
	SDstr[51]=0x20;
	SDstr[52]=0x20;
	SDstr[53]=0x20;
	SDstr[54]=0x46;
	SDstr[55]=0x41;
	SDstr[56]=0x54;
	SDstr[57]=0x31;
	SDstr[58]=0x36;
	SDstr[59]=0x20;
	SDstr[60]=0x20;
	SDstr[61]=0x20;
	SDstr[62]=0x33;
	SDstr[63]=0xc9;
	SD_WriteBlock(SDstr, 0);
	for(i=0;i<SD_BLOCK_SIZE;i++) SDstr[i]=0;
	SD_WriteBlock(SDstr, (uint32_t)(1*SD_BLOCK_SIZE));
	SD_WriteBlock(SDstr, (uint32_t)(2*SD_BLOCK_SIZE));
	SD_WriteBlock(SDstr, (uint32_t)(3*SD_BLOCK_SIZE));
	SD_WriteBlock(SDstr, (uint32_t)(4*SD_BLOCK_SIZE));
	SD_WriteBlock(SDstr, (uint32_t)(5*SD_BLOCK_SIZE));
	SerialPutString("Wrote block 0-5 (BPB)\r\n");
// end of BPB
// FAT (block 6-471)
	for(i=0;i<SD_BLOCK_SIZE;i++) SDstr[i]=0;
	SDstr[0]=0xF8;
	SDstr[1]=0xFF;
	SDstr[2]=0xFF;
	SDstr[3]=0xFF;
	SD_WriteBlock(SDstr, (uint32_t)(6*SD_BLOCK_SIZE));
	SerialPutString("Wrote block 6 (FAT)\r\n");
// end of FAT
// RDE (block 472-1d8)
	for(i=0;i<SD_BLOCK_SIZE;i++) SDstr[i]=0;
	SDstr[0]   ='S';
	SDstr[1]   ='D';
	SDstr[2]   ='C';
	SDstr[3]   ='A';
	SDstr[4]   ='R';
	SDstr[5]   ='D';
	SDstr[6]   =' ';
	SDstr[7]   =' ';
	SDstr[8]   =' ';
	SDstr[9]   =' ';
	SDstr[0xA] =' ';
	SDstr[0xB] =0x08; // volume name
	SDstr[32]  ='L';
	SDstr[32+1]='O';
	SDstr[32+2]='G';
	SDstr[32+3]=' ';
	SDstr[32+4]=' ';
	SDstr[32+5]=' ';
	SDstr[32+6]=' ';
	SDstr[32+7]=' ';
	SDstr[32+8]='T';
	SDstr[32+9]='X';
	SDstr[32+0xA]='T';
	SDstr[32+0xB]=0x20;// file
	SDstr[32+0xC]=0x18;// reserved
	SDstr[32+0x1A]=2; // start cluster [15:0]
	SDstr[32+0x1B]=0; // start cluster
	SDstr[32+0x1C]=0; // file syze BYTE [31:0]
	SDstr[32+0x1D]=0; // should be updated
	SDstr[32+0x1E]=0;
	SDstr[32+0x1F]=0;
	SD_WriteBlock(SDstr, (uint32_t)(472*SD_BLOCK_SIZE));
	SerialPutString("Wrote block 472 (RDE)\r\n");
// end of RDE
	return;
}

// SD card parameter calculation functions
extern u32 SD_size_byte; // 0-2GB
u16 SDstr_ptr(void)       {return (u16)( SD_size_byte%512);}     // 0-511
u32 SD_block(void)        {return (u32)((SD_size_byte>>9)+504);} // 504-4194303
u8  SD_block_cnt(void)    {return (u8) ((SD_size_byte>>9)%64);}  // 0-63
u16 SD_cluster(void)      {return (u16)(((SD_block()-504)>>6)+2);}     // 2-65535
u16 SD_FAT_block(void)    {return (u16)((SD_cluster()>>8)+6);}  // 6-471
u16 SD_FAT_block_ptr(void){return (u16)((SD_cluster()%256)*2);}  // 0-510

// wdata: write data    (If wdata is empty then flush buffer)
// v: "v" verbose mode
extern u8 SD_full; // SD card full flag
extern char UARTstrTX[80]; // USRT send buffer
void SD_write_sequencial_init(void) {
	SD_size_byte=0xFFFFFFFF; // 0-2GB
	SD_full=0; // SD card full flag
}
void SD_write_sequencial(u8* wdata, char v) {
	int i;
	u8* p;
	p=wdata;
	if(SD_full){
		SerialPutString("SD card full\r\n");
		return; // SD card full ?
	}
	// null wdata --> flush data
	if(wdata[0]==0){
		SerialPutString("\r\nSD flushing\r\n");
		// data flush
		if(SDstr_ptr()==511)SerialPutString("\r\n");// no data to be flushed
		else { // data to be flushed
			SD_WriteBlock(SDstr,(uint32_t)(SD_block()*SD_BLOCK_SIZE));
			sprintf(UARTstrTX,"EndCLUSTER %d EndBLOCK %d TotalSIZE %04x%04x[BYTE], ",SD_cluster(),(u16)SD_block(),(u16)(((SD_size_byte+1)&0xFFFF0000)>>16),(u16)((SD_size_byte+1)&0xFFFF));
			SerialPutString(UARTstrTX);
		}
		// update FAT
		SD_ReadBlock(SDstr,(uint32_t)((u32)SD_FAT_block()*SD_BLOCK_SIZE));
		SDstr[SD_FAT_block_ptr()]  =0xFF;
		SDstr[SD_FAT_block_ptr()+1]=0xFF;
		SD_WriteBlock(SDstr,(uint32_t)((u32)SD_FAT_block()*SD_BLOCK_SIZE));
		SerialPutString("FAT updated, ");
		// update RDE
		SD_size_byte++;
		SD_ReadBlock(SDstr,(uint32_t)(472*SD_BLOCK_SIZE));
		SDstr[32+0x1C]=(u8)( SD_size_byte&0xFF); // file syze BYTE [31:0]
		SDstr[32+0x1D]=(u8)((SD_size_byte&0xFF00)>>8);
		SDstr[32+0x1E]=(u8)((SD_size_byte&0xFF0000)>>16);
		SDstr[32+0x1F]=(u8)((SD_size_byte&0xFF000000)>>24);
		SD_WriteBlock(SDstr,(uint32_t)(472*SD_BLOCK_SIZE));
		SerialPutString("RDE updated\r\n");
	}
	// set data to SDstr
	for(i=0;i<strlen(wdata);i++) {
		SD_size_byte++; // file size
		SDstr[SDstr_ptr()]=*p++;
		if(SDstr_ptr()==511) {
			// write to SD card
			SD_WriteBlock(SDstr,(uint32_t)(SD_block()*SD_BLOCK_SIZE));
			// update FAT
			SD_ReadBlock(SDstr,(uint32_t)((u32)SD_FAT_block()*SD_BLOCK_SIZE));
			SDstr[SD_FAT_block_ptr()]  =(u8)((SD_cluster()+1)&0xFF);
			SDstr[SD_FAT_block_ptr()+1]=(u8)(((SD_cluster()+1)&0xFF00)>>8);
			SD_WriteBlock(SDstr,(uint32_t)((u32)SD_FAT_block()*SD_BLOCK_SIZE));
			// clr work area
			for(i=0;i<SD_BLOCK_SIZE;i++) SDstr[i]=0;
			// display
			if(SD_block_cnt()==0)	{
				sprintf(UARTstrTX,"\r\nCLUSTER%d",SD_cluster());
				SerialPutString(UARTstrTX);
				if(v=='v')SerialPutString("\r\n");
			}
			if(v=='v'){
				sprintf(UARTstrTX,"BLK[%d/%d]SizeB[%04x%04x]FAT[%d/%d] ",(u16)SD_block(),SD_block_cnt(),(u16)(((SD_size_byte+1)&0xFFFF0000)>>16),(u16)((SD_size_byte+1)&0xFFFF),SD_FAT_block(),SD_FAT_block_ptr());
				SerialPutString(UARTstrTX);
				if((SD_block_cnt()%2)==1)SerialPutString("\r\n");
			}
			else SerialPutString("-");
			// SD full ?
			if(SD_cluster()>=65530)	{
				SerialPutString("SD card is full\r\n");
				SD_full=1;
			}
		}
	}
	return;
}

uint8_t SD_WriteByte(uint8_t Data)
{
  // Wait until the transmit buffer is empty
  while (SPI_GetFlagStatus(SPI_FLAG_TXE) == RESET) {}
  SPI_SendData(Data);
  // Wait to receive a byte
  while (SPI_GetFlagStatus(SPI_FLAG_RXNE) == RESET) {}
  return SPI_ReceiveData();
}

uint8_t SD_ReadByte(void)
{
  uint8_t Data = 0;
  // Wait until the transmit buffer is empty
  while (SPI_GetFlagStatus(SPI_FLAG_TXE) == RESET) {}
  SPI_SendData(SD_DUMMY_BYTE);
  // Wait until a data is received
  while (SPI_GetFlagStatus(SPI_FLAG_RXNE) == RESET) {}
  Data = SPI_ReceiveData();
  return Data;
}

void SD_LowLevel_Init(void)
{
  // Enable SPI clock
  CLK_PeripheralClockConfig(SD_SPI_CLK, ENABLE);
  // Set the MOSI,MISO and SCK at high level
  GPIO_ExternalPullUpConfig(SD_SPI_SCK_GPIO_PORT, (GPIO_Pin_TypeDef)(SD_SPI_MISO_PIN | SD_SPI_MOSI_PIN | \
                            SD_SPI_SCK_PIN), ENABLE);
  // SD_SPI Configuration
  SPI_Init( SPI_FIRSTBIT_MSB, SPI_BAUDRATEPRESCALER_4, SPI_MODE_MASTER,
           SPI_CLOCKPOLARITY_HIGH, SPI_CLOCKPHASE_2EDGE, SPI_DATADIRECTION_2LINES_FULLDUPLEX,
           SPI_NSS_SOFT, 0x07);
  SPI_Cmd( ENABLE);  // SD_SPI enable
  // Set MSD ChipSelect pin in Output push-pull high level
  GPIO_Init(SD_CS_GPIO_PORT, SD_CS_PIN, GPIO_MODE_OUT_PP_HIGH_SLOW);
}

uint8_t SD_Init(void)
{
  uint32_t i = 0;
  SD_LowLevel_Init();  // Initialize SD_SPI
  SD_CS_LOW();  // SD chip select assert
  for (i = 0; i <= 9; i++)  // wait 80 clocks cycles
  {
    SD_WriteByte(SD_DUMMY_BYTE); // Send dummy byte 0xFF
  }
  // SD initialized and set to SPI mode properly */
  return (SD_GoIdleState());
}

uint8_t SD_GoIdleState(void)
{
  SD_CS_LOW(); // SD CS assert
  // Send CMD0 (SD_CMD_GO_IDLE_STATE) to put SD in SPI mode
  SD_SendCmd(SD_CMD_GO_IDLE_STATE, 0, 0x95);
  // Wait for In Idle State Response (R1 Format) equal to 0x01
  if (SD_GetResponse(SD_IN_IDLE_STATE))
  {
    // No Idle State Response: return response failure
    return SD_RESPONSE_FAILURE;
  }
  //---------Activates the card initialization process-----------
  do
  {
    SD_CS_HIGH(); // SD CS deassert
    SD_WriteByte(SD_DUMMY_BYTE); // dummy 8 clocks
    SD_CS_LOW(); // SD CS assert
    // Send CMD1 (Activates the card process) until response equal to 0x0
    SD_SendCmd(SD_CMD_SEND_OP_COND, 0, 0xFF);
    // Wait for no error Response (R1 Format) equal to 0x00
  }
  while (SD_GetResponse(SD_RESPONSE_NO_ERROR));
  SD_CS_HIGH(); // SD CS deassert
  SD_WriteByte(SD_DUMMY_BYTE); // dummy 8 clocks
  return SD_RESPONSE_NO_ERROR;
}

void SD_SendCmd(uint8_t Cmd, uint32_t Arg, uint8_t Crc)
{
  uint32_t i = 0x00;
  uint8_t Frame[6];
  Frame[0] = (uint8_t)(Cmd | 0x40); // Construct byte 1
  Frame[1] = (uint8_t)(Arg >> 24);  // Construct byte 2
  Frame[2] = (uint8_t)(Arg >> 16);  // Construct byte 3
  Frame[3] = (uint8_t)(Arg >> 8);   // Construct byte 4
  Frame[4] = (uint8_t)(Arg);        // Construct byte 5
  Frame[5] = (Crc);                 // Construct CRC: byte 6
  for (i = 0; i < 6; i++)
  {
    SD_WriteByte(Frame[i]); // Send the Cmd bytes
  }
}

uint8_t SD_Detect(void)
{
  __IO uint8_t status = SD_PRESENT;
  // Check GPIO to detect SD
  if (GPIO_ReadInputData(SD_DETECT_GPIO_PORT) & SD_DETECT_PIN)
  {
    status = SD_NOT_PRESENT;
  }
  return status;
}

uint8_t SD_ReadBlock(uint8_t* pBuffer, uint32_t ReadAddr)
{
  uint32_t i = 0;
  uint8_t rvalue = SD_RESPONSE_FAILURE;
	GPIO_WriteLow(GPIOD, GPIO_PIN_0);  // LED on
  SD_CS_LOW(); // SD CS assert
  // Send CMD17 (SD_CMD_READ_SINGLE_BLOCK) to read one block
  SD_SendCmd(SD_CMD_READ_SINGLE_BLOCK, ReadAddr, 0xFF);
  // Check if the SD acknowledged the read block command: R1 response (0x00: no errors) */
  if (!SD_GetResponse(SD_RESPONSE_NO_ERROR))
  {
    // wait data token(0xFE) to signify the start of the data
    if (!SD_GetResponse(SD_START_DATA_SINGLE_BLOCK_READ))
    {
      // Read 512 BYTE data
      for (i = 0; i < SD_BLOCK_SIZE; i++)
      {
        *pBuffer = SD_ReadByte();
        pBuffer++;
      }
      // Get CRC bytes (not really needed by us, but required by SD)
      SD_ReadByte();
      SD_ReadByte();
      rvalue = SD_RESPONSE_NO_ERROR;
    }
  }
  SD_CS_HIGH(); // SD CS deassert
  SD_WriteByte(SD_DUMMY_BYTE); // dummy 8 clocks
	GPIO_WriteHigh(GPIOD, GPIO_PIN_0);  // LED off
  return rvalue;
}

uint8_t SD_WriteBlock(uint8_t* pBuffer, uint32_t WriteAddr)
{
  uint32_t i = 0;
  uint8_t rvalue = SD_RESPONSE_FAILURE;
	GPIO_WriteLow(GPIOD, GPIO_PIN_0);  // LED on
  SD_CS_LOW(); // SD CS assert
  // Send CMD24 (SD_CMD_WRITE_SINGLE_BLOCK) to write a block
  SD_SendCmd(SD_CMD_WRITE_SINGLE_BLOCK, WriteAddr, 0xFF);
  // Check command response R1 (0x00: no errors)
  if (!SD_GetResponse(SD_RESPONSE_NO_ERROR))
  {
    SD_WriteByte(SD_DUMMY_BYTE); // Send a dummy byte
    SD_WriteByte(0xFE);// Send data token
    for (i = 0; i < SD_BLOCK_SIZE; i++) // Send data
    {
      SD_WriteByte(*pBuffer);
      pBuffer++;
    }
    SD_ReadByte(); // Send CRC
    SD_ReadByte();
    // Read data response
    if (SD_GetDataResponse() == SD_DATA_OK)
    {
      rvalue = SD_RESPONSE_NO_ERROR;
    }
  }
  SD_CS_HIGH(); // SD CS deassert
  SD_WriteByte(SD_DUMMY_BYTE); // dummy 8 clocks
	GPIO_WriteHigh(GPIOD, GPIO_PIN_0);  // LED off
  return rvalue;
}

uint8_t SD_GetDataResponse(void)
{
  uint32_t i = 0;
  uint8_t response = 0, rvalue = 0;

  while (i <= 64)
  {
    response = SD_ReadByte(); // Read response
    response &= 0x1F; // Mask unused bits
    switch (response)
    {
      case SD_DATA_OK:
      {
        rvalue = SD_DATA_OK;
        break;
      }
      case SD_DATA_CRC_ERROR:   return SD_DATA_CRC_ERROR;
      case SD_DATA_WRITE_ERROR: return SD_DATA_WRITE_ERROR;
      default:
      {
        rvalue = SD_DATA_OTHER_ERROR;
        break;
      }
    }
    // Exit loop in case of data ok
    if (rvalue == SD_DATA_OK) break;
    i++;
  }
  while (SD_ReadByte() == 0);  // Wait null data
  return response;
}

uint8_t SD_GetResponse(uint8_t Response)
{
  uint32_t Count = 0xFFF;
  // Check if response is got or a timeout is happen
  while ((SD_ReadByte() != Response) && Count)
  {
    Count--;
  }
  if (Count == 0) // timeout occured
  {
    return SD_RESPONSE_FAILURE;
  }
  else // Good response
  {
    return SD_RESPONSE_NO_ERROR;
  }
}

void SD_LowLevel_DeInit(void)
{
  SPI_Cmd(DISABLE); // SD_SPI disable
  // SD_SPI Peripheral clock disable
  CLK_PeripheralClockConfig(SD_SPI_CLK, DISABLE);
  // Configure SD_SPI pins: SCK
  GPIO_Init(SD_SPI_SCK_GPIO_PORT, SD_SPI_SCK_PIN, GPIO_MODE_IN_FL_NO_IT);
  // Configure SD_SPI pins: MISO
  GPIO_Init(SD_SPI_MISO_GPIO_PORT, SD_SPI_MISO_PIN, GPIO_MODE_IN_FL_NO_IT);
  // Configure SD_SPI pins: MOSI
  GPIO_Init(SD_SPI_MOSI_GPIO_PORT, SD_SPI_MOSI_PIN, GPIO_MODE_IN_FL_NO_IT);
  // Configure SD_SPI_CS_PIN pin: SD Card CS pin
  GPIO_Init(SD_CS_GPIO_PORT, SD_CS_PIN, GPIO_MODE_IN_FL_NO_IT);
  // Configure SD_SPI_DETECT_PIN pin: SD Card detect pin
  GPIO_Init(SD_DETECT_GPIO_PORT, SD_DETECT_PIN, GPIO_MODE_IN_FL_NO_IT);
}

void SD_DeInit(void)
{
  SD_LowLevel_DeInit();
}
