#include "stm8s.h"

// hardware definitions
#define SD_SPI_CLK                CLK_PERIPHERAL_SPI
#define SD_SPI_SCK_GPIO_PORT      GPIOC           // GPIOC
#define SD_SPI_SCK_PIN            GPIO_PIN_5      // PC5
#define SD_SPI_MISO_GPIO_PORT     GPIOC           // GPIOC
#define SD_SPI_MISO_PIN           GPIO_PIN_7      // PC7
#define SD_SPI_MOSI_GPIO_PORT     GPIOC           // GPIOC
#define SD_SPI_MOSI_PIN           GPIO_PIN_6      // PC6
#define SD_CS_GPIO_PORT           GPIOE           // GPIOE
#define SD_CS_PIN                 GPIO_PIN_5      // PE5
#define SD_DETECT_GPIO_PORT       GPIOG           // GPIOG
#define SD_DETECT_PIN             GPIO_PIN_0      // PG0
    
// SD card parameters
#define SD_PRESENT        ((uint8_t)0x01)
#define SD_NOT_PRESENT    ((uint8_t)0x00)
#define SD_BLOCK_SIZE    0x200
#define SD_DUMMY_BYTE   0xFF

// SD work area
extern uint8_t SDFAT[SD_BLOCK_SIZE]; // FAT work area
extern uint8_t SDstr[SD_BLOCK_SIZE]; // SD card read/write buffer 512bytes

// SD commands
#define SD_CMD_GO_IDLE_STATE          0   /*!< CMD0 = 0x40 */
#define SD_CMD_SEND_OP_COND           1   /*!< CMD1 = 0x41 */
#define SD_CMD_SEND_CSD               9   /*!< CMD9 = 0x49 */
#define SD_CMD_SEND_CID               10  /*!< CMD10 = 0x4A */
#define SD_CMD_STOP_TRANSMISSION      12  /*!< CMD12 = 0x4C */
#define SD_CMD_SEND_STATUS            13  /*!< CMD13 = 0x4D */
#define SD_CMD_SET_BLOCKLEN           16  /*!< CMD16 = 0x50 */
#define SD_CMD_READ_SINGLE_BLOCK      17  /*!< CMD17 = 0x51 */
#define SD_CMD_READ_MULT_BLOCK        18  /*!< CMD18 = 0x52 */
#define SD_CMD_SET_BLOCK_COUNT        23  /*!< CMD23 = 0x57 */
#define SD_CMD_WRITE_SINGLE_BLOCK     24  /*!< CMD24 = 0x58 */
#define SD_CMD_WRITE_MULT_BLOCK       25  /*!< CMD25 = 0x59 */
#define SD_CMD_PROG_CSD               27  /*!< CMD27 = 0x5B */
#define SD_CMD_SET_WRITE_PROT         28  /*!< CMD28 = 0x5C */
#define SD_CMD_CLR_WRITE_PROT         29  /*!< CMD29 = 0x5D */
#define SD_CMD_SEND_WRITE_PROT        30  /*!< CMD30 = 0x5E */
#define SD_CMD_SD_ERASE_GRP_START     32  /*!< CMD32 = 0x60 */
#define SD_CMD_SD_ERASE_GRP_END       33  /*!< CMD33 = 0x61 */
#define SD_CMD_UNTAG_SECTOR           34  /*!< CMD34 = 0x62 */
#define SD_CMD_ERASE_GRP_START        35  /*!< CMD35 = 0x63 */
#define SD_CMD_ERASE_GRP_END          36  /*!< CMD36 = 0x64 */
#define SD_CMD_UNTAG_ERASE_GROUP      37  /*!< CMD37 = 0x65 */
#define SD_CMD_ERASE                  38  /*!< CMD38 = 0x66 */

// Start Data tokens
#define SD_START_DATA_SINGLE_BLOCK_READ    0xFE  /*!< Data token start byte, Start Single Block Read */
#define SD_START_DATA_MULTIPLE_BLOCK_READ  0xFE  /*!< Data token start byte, Start Multiple Block Read */
#define SD_START_DATA_SINGLE_BLOCK_WRITE   0xFE  /*!< Data token start byte, Start Single Block Write */
#define SD_START_DATA_MULTIPLE_BLOCK_WRITE 0xFD  /*!< Data token start byte, Start Multiple Block Write */
#define SD_STOP_DATA_MULTIPLE_BLOCK_WRITE  0xFD  /*!< Data toke stop byte, Stop Multiple Block Write */

// SD responses and error flags
#define  SD_RESPONSE_NO_ERROR         0x00
#define  SD_IN_IDLE_STATE             0x01
#define  SD_ERASE_RESET               0x02
#define  SD_ILLEGAL_COMMAND           0x04
#define  SD_COM_CRC_ERROR             0x08
#define  SD_ERASE_SEQUENCE_ERROR      0x10
#define  SD_ADDRESS_ERROR             0x20
#define  SD_PARAMETER_ERROR           0x40
#define  SD_RESPONSE_FAILURE          0xFF

// Data response error flags
#define  SD_DATA_OK                   0x05
#define  SD_DATA_CRC_ERROR            0x0B
#define  SD_DATA_WRITE_ERROR          0x0D
#define  SD_DATA_OTHER_ERROR          0xFF

/*----- FAT16 function -----*/
extern u32 SD_size_byte; // 0-2GB
extern u8 SD_full; // SD card full flag
u16 SDstr_ptr(void);        // 0-511
u32 SD_block(void);         // 504-4194303
u8  SD_block_cnt(void);     // 0-63
u16 SD_cluster(void);       // 2-65535
u16 SD_FAT_block(void);     // 6-471
u16 SD_FAT_block_ptr(void); // 0-510
void SD_write_sequencial_init(void); // initialize
void SD_write_sequencial(u8* wdata, char v); // write
void SD_format_2GB(void);

/*----- High layer function -----*/
void SD_DeInit(void);  
uint8_t SD_Init(void);
uint8_t SD_Detect(void);
uint8_t SD_ReadBlock(uint8_t* pBuffer, uint32_t ReadAddr);
uint8_t SD_WriteBlock(uint8_t* pBuffer, uint32_t WriteAddr);
/*----- Medium layer function -----*/
void SD_SendCmd(uint8_t Cmd, uint32_t Arg, uint8_t Crc);
uint8_t SD_GetResponse(uint8_t Response);
uint8_t SD_GetDataResponse(void);
uint8_t SD_GoIdleState(void);
/*----- Low layer function -----*/
uint8_t SD_WriteByte(uint8_t byte);
uint8_t SD_ReadByte(void);
void SD_LowLevel_DeInit(void);
void SD_LowLevel_Init(void);
#define SD_CS_LOW()     GPIO_WriteLow(SD_CS_GPIO_PORT, SD_CS_PIN)
#define SD_CS_HIGH()    GPIO_WriteHigh(SD_CS_GPIO_PORT, SD_CS_PIN)
