#include "stdio.h"
#include "stdlib.h"
#include "stm8s.h"
#include "usrlib-uart.h"

void SerialPutChar(char c)
{
    UART2_SendData8(c);
    while ((UART2->SR & UART2_SR_TXE ) != UART2_SR_TXE );
}

void SerialPutString(char *s)
{
	while (*s != '\0')
	{
		SerialPutChar(*s);
		s ++;
	}
}

// divide a line into words
//  input:
//		tmp = "abc def ghi jklm"
//  output:
//		Field[0] = "abc"
//  	Field[1] = "def"
//  	Field[2] = "ghi"
//  	Field[3] = "jklm"
char Field[FIELD_NUM][FIELD_CHAR];
int separate_line(char *tmp){
  int  p,f,fp;
  p=0;
  while(tmp[p]<=32){   // neglection top space
    if(tmp[p]==0) {Field[0][0]=0; return(0);}
		else{
			p++;
			if(p==FIELD_NUM*FIELD_CHAR-1){
				Field[0][0]=0;
				return(0);
			}
		}
  }
  f=0; fp=0;
  while(1){
    if(tmp[p]==0){  // end of line
	    Field[f][fp]=0;
	    if(fp==0) return(f);
	    else return(f+1);
	  }
    else if(tmp[p]<=32){  // space found
	    if(tmp[p-1]>32){  // there is a character
	      Field[f][fp]=0;
	      f++; fp=0;   // prepair next word
				if(f==FIELD_NUM){
					Field[0][0]=0;
					return(0);
				}
	      p++;
				if(p==FIELD_NUM*FIELD_CHAR-1){
					Field[0][0]=0;
					return(0);
				}
	    }
			else{   // nothing character
				p++;
				if(p==FIELD_NUM*FIELD_CHAR-1){
					Field[0][0]=0;
					return(0);
				}
			}
	  }
    else{  // found a character
	    Field[f][fp]=tmp[p];
	    fp++;
			if(fp==FIELD_CHAR){
				Field[0][0]=0;
				return(0);
			}
	    p++;
			if(p==FIELD_NUM*FIELD_CHAR-1){
				Field[0][0]=0; 
				return(0);
			}
	  }
  }
}

// UART RX interrupt routine
// interrupt each characters
char command[FIELD_NUM*FIELD_CHAR];
u8 UART_STR_EXIST = 0;
int bytes_read = 0;

void UART2RX_isr(void) interrupt 21
{
	if(UART2_GetITStatus(UART2_IT_RXNE)==SET)
	{
		char c;
		if(UART_STR_EXIST==1) return;
		c = UART2_ReceiveData8();
		if(c == '\r'){  // CR
			command[bytes_read] = '\0';
			bytes_read = 0;
			UART_STR_EXIST = 1;   // command string exist flag
			return;
		}
		else if(c == '\b'){  // BS
			if (bytes_read > 0){
				SerialPutString("\b \b");
				bytes_read--;
			}
			return;
		}
		else if(bytes_read >= FIELD_NUM*FIELD_CHAR ){
			SerialPutString("Command too long\r\n");
			bytes_read = 0;
			return;
		}
		else if(c >= 0x20 && c <= 0x7E){
			command[bytes_read++] = c;
			SerialPutChar(c);
		}
	}
}
